# -*- coding: utf-8 -*-
from ._datetimes import is_date_or_time
from ._numbers import is_number


QUOTE = b'"'
ALT_QUOTE = b"'"

FORMATTING_CHARS = {
    b'n': b'\n',
    b't': b'\t',
    b'f': b'\f',
    b'v': b'\v',
    b'\\': b'\\',
}

RESTRICTED_SEQ = [
    b'/*', b'*/',  # Comments
]

RESTRICTED_CHARS = set(
    b' \r\n\t\v\f' +        # Whitespace
    b'&<>\'{},[]=!#()%";|'  # Restricted chars
)

RESERVED_KEYWORDS = set([
    b'Null', b'NULL',
    b'End', b'END',
    b'TRUE', b'True', b'true',
    b'FALSE', b'False', b'false',
    b'Group', b'GROUP', b'BEGIN_GROUP',
    b'End_Group', b'END_GROUP',
    b'Object', b'OBJECT', b'BEGIN_OBJECT',
    b'End_Object', b'END_OBJECT',
    b'End', 'END',
])


def escape_quote(quote, value):
    return quote + value.replace(quote, b'\\' + quote) + quote


def quote_string(value):
    for escape, char in FORMATTING_CHARS.items():
        value = value.replace(char, b'\\' + escape)

    if QUOTE in value and ALT_QUOTE not in value:
        return escape_quote(ALT_QUOTE, value)

    return escape_quote(QUOTE, value)


def needs_quotes(value):
    if not value:
        return True

    if is_number(value):
        return True

    if is_date_or_time(value):
        return True

    if value in RESERVED_KEYWORDS:
        return True

    for char in value:
        if char in RESTRICTED_CHARS:
            return True

    for seq in RESTRICTED_SEQ:
        if seq in value:
            return True

    return False



# -*- coding: utf-8 -*-
import pytest
import six
import pvl


class DictLike(object):
    def keys(self):
        return ['a', 'b', 'a']

    def __getitem__(self, key):
        return 42


def test_empty():
    module = pvl.PVLModule()

    assert len(module) == 0
    assert module.get('c', 42) == 42

    with pytest.raises(KeyError):
        module['c']


def test_list_creation():
    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    assert len(module) == 3
    assert module['a'] == 1
    assert module['b'] == 2
    assert module.getlist('a') == [1, 3]

    with pytest.raises(KeyError):
        module['c']

    assert module.get('c', 42) == 42

    with pytest.raises(TypeError):
        pvl.PVLModule([], [])

    module = pvl.PVLModule(DictLike())
    assert len(module) == 3
    assert module['a'] == 42
    assert module['b'] == 42
    assert module.getlist('a') == [42, 42]

    with pytest.raises(KeyError):
        module['c']


def test_dict_creation():
    module = pvl.PVLModule({'a': 1, 'b': 2})

    assert module['a'] == 1
    assert module['b'] == 2
    assert len(module) == 2

    with pytest.raises(KeyError):
        module['c']

    assert module.get('c', 42) == 42


def test_keyword_creation():
    module = pvl.PVLModule(a=1, b=2)

    assert module['a'] == 1
    assert module['b'] == 2
    assert len(module) == 2

    with pytest.raises(KeyError):
        module['c']

    assert module.get('c', 42) == 42


def test_key_access():
    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    assert module['a'] == 1
    assert module['b'] == 2

    with pytest.raises(KeyError):
        module['c']


def test_index_access():
    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    assert module[0] == ('a', 1)
    assert module[1] == ('b', 2)
    assert module[2] == ('a', 3)

    with pytest.raises(IndexError):
        module[3]


def test_slice_access():
    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    assert module[0:3] == [('a', 1), ('b', 2), ('a', 3)]
    assert module[1:] == [('b', 2), ('a', 3)]
    assert module[:-1] == [('a', 1), ('b', 2)]


def test_set():
    module = pvl.PVLModule()
    module['a'] = 1
    module['b'] = 2
    module['a'] = 3

    assert module['a'] == 3
    assert module['b'] == 2
    assert module.getlist('a') == [3]
    assert len(module) == 2

    with pytest.raises(KeyError):
        module['c']

    assert module.get('c', 42) == 42
    assert list(module) == [('a', 3), ('b', 2)]


def test_delete():
    module = pvl.PVLModule(a=1, b=2)

    assert len(module) == 2
    assert module['a'] == 1

    del module['a']
    assert len(module) == 1

    with pytest.raises(KeyError):
        module['a']

    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    assert len(module) == 3
    assert module['a'] == 1

    del module['a']
    assert len(module) == 1

    with pytest.raises(KeyError):
        module['a']

    with pytest.raises(KeyError):
        del module['c']


def test_clear():
    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    assert len(module) == 3
    assert module['a'] == 1

    module.clear()
    assert len(module) == 0
    assert module.getlist('a') == []

    with pytest.raises(KeyError):
        module['a']

    with pytest.raises(KeyError):
        module['b']

    module['a'] = 42
    assert len(module) == 1
    assert module['a'] == 42


def test_discard():
    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    assert len(module) == 3
    assert module['a'] == 1

    module.discard('a')
    assert len(module) == 1
    assert module.getlist('a') == []

    with pytest.raises(KeyError):
        module['a']

    assert module['b'] == 2
    module.discard('b')

    assert len(module) == 0

    with pytest.raises(KeyError):
        module['b']

    module.discard('c')
    assert len(module) == 0


def test_pop():
    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    assert len(module) == 3
    assert module.pop('a') == 1
    assert len(module) == 1

    with pytest.raises(KeyError):
        module['a']

    with pytest.raises(KeyError):
        module.pop('a')

    assert module.pop('a', 42) == 42

    assert module.pop('b') == 2
    assert len(module) == 0

    with pytest.raises(KeyError):
        module.pop('b')

    with pytest.raises(KeyError):
        module['b']

    assert module.pop('b', 42) == 42

    with pytest.raises(KeyError):
        module.pop('c')

    assert module.pop('c', 42) == 42


def test_popitem():
    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    assert len(module) == 3

    assert module.popitem() == ('a', 3)
    assert len(module) == 2

    assert module.popitem() == ('b', 2)
    assert len(module) == 1

    assert module.popitem() == ('a', 1)
    assert len(module) == 0

    with pytest.raises(KeyError):
        module.popitem()


def test_update():
    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    module.update({'a': 42, 'c': 7})
    assert len(module) == 3
    assert module['a'] == 42
    assert module['b'] == 2
    assert module['c'] == 7

    module.update()
    assert len(module) == 3
    assert module['a'] == 42
    assert module['b'] == 2
    assert module['c'] == 7


def test_append():
    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    module.append('a', 42)
    assert len(module) == 4
    assert module['a'] == 1
    assert module.getlist('a') == [1, 3, 42]

    module.append('c', 43)
    assert len(module) == 5
    assert module['c'] == 43
    assert module.getlist('c') == [43]


def test_len():
    module = pvl.PVLModule()
    assert len(module) == 0

    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])
    assert len(module) == 3


def test_repr():
    module = pvl.PVLModule()
    assert isinstance(repr(module), str)
    assert repr(module) == 'PVLModule([])'

    module = pvl.PVLModule(a=1)
    assert isinstance(repr(module), str)


@pytest.mark.skipif(six.PY3, reason='requires python2')
def test_py2_items():
    module = pvl.PVLModule()

    assert isinstance(module.items(), list)
    assert module.items() == []

    assert isinstance(module.keys(), list)
    assert module.keys() == []

    assert isinstance(module.values(), list)
    assert module.values() == []

    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    assert isinstance(module.items(), list)
    assert module.items() == [('a', 1), ('b', 2), ('a', 3)]

    assert isinstance(module.keys(), list)
    assert module.keys() == ['a', 'b', 'a']

    assert isinstance(module.values(), list)
    assert module.values() == [1, 2, 3]


if six.PY3:
    def iteritems(module):
        return module.items()

    def iterkeys(module):
        return module.keys()

    def itervalues(module):
        return module.values()

else:
    def iteritems(module):
        return module.iteritems()

    def iterkeys(module):
        return module.iterkeys()

    def itervalues(module):
        return module.itervalues()


def test_iterators():
    module = pvl.PVLModule()

    assert isinstance(iteritems(module), pvl._collections.MappingView)
    assert list(iteritems(module)) == []
    assert len(iteritems(module)) == 0
    assert isinstance(repr(iteritems(module)), str)
    assert ('a', 1) not in iteritems(module)

    assert isinstance(iterkeys(module), pvl._collections.MappingView)
    assert list(iterkeys(module)) == []
    assert len(iterkeys(module)) == 0
    assert isinstance(repr(iterkeys(module)), str)
    assert 'a' not in iterkeys(module)

    assert isinstance(itervalues(module), pvl._collections.MappingView)
    assert list(itervalues(module)) == []
    assert len(itervalues(module)) == 0
    assert isinstance(repr(itervalues(module)), str)
    assert 1 not in itervalues(module)

    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    assert isinstance(iteritems(module), pvl._collections.MappingView)
    assert list(iteritems(module)) == [('a', 1), ('b', 2), ('a', 3)]
    assert len(iteritems(module)) == 3
    assert isinstance(repr(iteritems(module)), str)
    assert ('a', 1) in iteritems(module)
    assert ('b', 2) in iteritems(module)
    assert ('a', 3) in iteritems(module)
    assert ('c', 4) not in iteritems(module)

    assert isinstance(iterkeys(module), pvl._collections.MappingView)
    assert list(iterkeys(module)) == ['a', 'b', 'a']
    assert len(iterkeys(module)) == 3
    assert isinstance(repr(iterkeys(module)), str)
    assert 'a' in iterkeys(module)
    assert 'b' in iterkeys(module)
    assert 'c' not in iterkeys(module)

    assert isinstance(itervalues(module), pvl._collections.MappingView)
    assert list(itervalues(module)) == [1, 2, 3]
    assert len(itervalues(module)) == 3
    assert isinstance(repr(itervalues(module)), str)
    assert 1 in itervalues(module)
    assert 2 in itervalues(module)
    assert 3 in itervalues(module)
    assert 4 not in itervalues(module)


def test_equlity():
    assert not pvl.PVLModule()
    assert not pvl.PVLGroup()
    assert not pvl.PVLObject()

    assert not not pvl.PVLModule(a=1)
    assert not not pvl.PVLGroup(a=1)
    assert not not pvl.PVLObject(a=1)

    assert pvl.PVLModule() == pvl.PVLModule()
    assert pvl.PVLModule() != pvl.PVLGroup()
    assert pvl.PVLModule() != pvl.PVLObject()

    assert pvl.PVLGroup() != pvl.PVLModule()
    assert pvl.PVLGroup() == pvl.PVLGroup()
    assert pvl.PVLGroup() != pvl.PVLObject()

    assert pvl.PVLObject() != pvl.PVLModule()
    assert pvl.PVLObject() != pvl.PVLGroup()
    assert pvl.PVLObject() == pvl.PVLObject()

    assert pvl.PVLModule() != pvl.PVLModule(a=1)
    assert pvl.PVLModule(a=1) == pvl.PVLModule(a=1)
    assert pvl.PVLModule(a=1) == pvl.PVLModule([('a', 1)])
    assert pvl.PVLModule(a=1) == pvl.PVLModule({'a': 1})
    assert pvl.PVLModule(a=1) != pvl.PVLModule(b=1)
    assert pvl.PVLModule(a=1) != pvl.PVLModule(a=2)


def test_copy():
    module = pvl.PVLModule()
    copy = module.copy()

    assert module == copy
    assert module is not copy

    module['c'] = 42
    assert module != copy

    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])
    copy = module.copy()

    assert module == copy
    assert module is not copy

    module['c'] = 42
    assert module != copy


def test_conversion():
    module = pvl.PVLModule([
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ])

    expected_dict = {
        'a': [1, 3],
        'b': [2],
    }

    expected_list = [
        ('a', 1),
        ('b', 2),
        ('a', 3),
    ]

    assert dict(module) == expected_dict
    assert list(module) == expected_list



# -*- coding: utf-8 -*-
import datetime
import six
from ._collections import PVLGroup, Units
from ._strings import needs_quotes, quote_string


class PVLEncoder(object):
    begin_group = b'BEGIN_GROUP'
    end_group = b'END_GROUP'
    begin_object = b'BEGIN_OBJECT'
    end_object = b'END_OBJECT'
    end_statement = b'END'
    null = b'NULL'
    true = b'TRUE'
    false = b'FALSE'
    assignment = b' = '
    indentation = b'  '
    newline = b'\r\n'

    def indent(self, level, stream):
        stream.write(level * self.indentation)

    def encode(self, module, stream):
        self.encode_block(module, 0, stream)
        stream.write(self.end_statement)

    def encode_block(self, block, level, stream):
        for key, value in six.iteritems(block):
            self.encode_statement(key, value, level, stream)

    def encode_statement(self, key, value, level, stream):
        if isinstance(value, PVLGroup):
            return self.encode_group(key, value, level, stream)

        if isinstance(value, dict):
            return self.encode_object(key, value, level, stream)

        self.encode_assignment(key, value, level, stream)

    def encode_group(self, key, value, level, stream):
        self.encode_group_begin(key, value, level, stream)
        self.encode_block(value, level + 1, stream)
        self.encode_group_end(key, value, level, stream)

    def encode_group_begin(self, key, value, level, stream):
        self.encode_raw_assignment(
            key=self.begin_group,
            value=key.encode('utf-8'),
            level=level,
            stream=stream
        )

    def encode_group_end(self, key, value, level, stream):
        self.encode_raw_assignment(
            key=self.end_group,
            value=key.encode('utf-8'),
            level=level,
            stream=stream
        )

    def encode_object(self, key, value, level, stream):
        self.encode_object_begin(key, value, level, stream)
        self.encode_block(value, level + 1, stream)
        self.encode_object_end(key, value, level, stream)

    def encode_object_begin(self, key, value, level, stream):
        self.encode_raw_assignment(
            key=self.begin_object,
            value=key.encode('utf-8'),
            level=level,
            stream=stream
        )

    def encode_object_end(self, key, value, level, stream):
        self.encode_raw_assignment(
            key=self.end_object,
            value=key.encode('utf-8'),
            level=level,
            stream=stream
        )

    def encode_assignment(self, key, value, level, stream):
        self.encode_raw_assignment(
            key=key.encode('utf-8'),
            value=self.encode_value(value),
            level=level,
            stream=stream
        )

    def encode_raw_assignment(self, key, value, level, stream):
        self.indent(level, stream)
        stream.write(key)
        stream.write(self.assignment)
        stream.write(value)
        stream.write(self.newline)

    def encode_value(self, value):
        if isinstance(value, Units):
            units = self.encode_units(value.units)
            value = self.encode_simple_value(value.value)
            return value + b' ' + units
        return self.encode_simple_value(value)

    def encode_simple_value(self, value):
        if isinstance(value, six.string_types):
            return self.encode_string(value)

        if value is None:
            return self.encode_null(value)

        if isinstance(value, bool):
            return self.encode_bool(value)

        if isinstance(value, (int, float)):
            return self.encode_number(value)

        if isinstance(value, list):
            return self.encode_list(value)

        if isinstance(value, set):
            return self.encode_set(value)

        if isinstance(value, datetime.datetime):
            return self.encode_datetime(value)

        if isinstance(value, datetime.date):
            return self.encode_date(value)

        if isinstance(value, datetime.time):
            return self.encode_time(value)

        return self.default(value)

    def encode_units(self, value):
        return b'<' + value.encode('utf-8') + b'>'

    def encode_null(self, value):
        return self.null

    def encode_number(self, value):
        return repr(value).encode('utf-8')

    def encode_string(self, value):
        value = value.encode('utf-8')
        if needs_quotes(value):
            return quote_string(value)
        return value

    def encode_bool(self, value):
        if value:
            return self.true
        return self.false

    def encode_date(self, value):
        date = u'%04d-%02d-%02d' % (value.year, value.month, value.day)
        return date.encode('utf-8')

    def encode_tz(self, offset):
        hours = int(offset.seconds / 3600) + (offset.days * 24)
        return u'%+d' % hours

    def encode_time(self, value):
        if value.microsecond:
            second = u'%02d.%06d' % (value.second, value.microsecond)
        else:
            second = u'%02d' % value.second

        if value.utcoffset() is not None:
            second += self.encode_tz(value.utcoffset())

        time = u'%02d:%02d:%s' % (value.hour, value.minute, second)
        return time.encode('utf-8')

    def encode_datetime(self, value):
        date = self.encode_date(value)
        time = self.encode_time(value)
        return date + b'T' + time

    def encode_sequence(self, values):
        return b', '.join([self.encode_value(v) for v in values])

    def encode_list(self, value):
        return b'(' + self.encode_sequence(value) + b')'

    def encode_set(self, value):
        return b'{' + self.encode_sequence(value) + b'}'

    def default(self, value):
        raise TypeError(repr(value) + " is not serializable")


class IsisCubeLabelEncoder(PVLEncoder):
    begin_group = b'Group'
    end_group = b'End_Group'
    begin_object = b'Object'
    end_object = b'End_Object'
    end_statement = b'End'

    def encode_group_end(self, key, value, level, stream):
        self.indent(level, stream)
        stream.write(self.end_group)
        stream.write(self.newline)

    def encode_object_end(self, key, value, level, stream):
        self.indent(level, stream)
        stream.write(self.end_object)
        stream.write(self.newline)


class PDSLabelEncoder(PVLEncoder):
    begin_group = b'GROUP'
    begin_object = b'OBJECT'

    def _detect_assignment_col(self, block, indent=0):
        if not block:
            return 0
        block_items = six.iteritems(block)
        return max(self._key_length(k, v, indent) for k, v in block_items)

    def _key_length(self, key, value, indent):
        length = indent + len(key)

        if isinstance(value, dict):
            indent += len(self.indentation)
            return max(length, self._detect_assignment_col(value, indent))

        return length

    def encode(self, module, stream):
        self.assignment_col = self._detect_assignment_col(module)
        super(PDSLabelEncoder, self).encode(module, stream)

    def encode_raw_assignment(self, key, value, level, stream):
        indented_key = (level * self.indentation) + key
        stream.write(indented_key.ljust(self.assignment_col))
        stream.write(self.assignment)
        stream.write(value)
        stream.write(self.newline)
